use std::sync::mpsc::Receiver;

use color_eyre::{eyre, Result};
use mlua::{Function, IntoLua as _, Lua, LuaSerdeExt};
use tokio::sync::mpsc::UnboundedSender;

use crate::types::{ApiTask, Event, Message, NtfyMessage};

#[tracing::instrument(skip_all)]
pub fn worker(
    config: String,
    event_rx: Receiver<Event>,
    api_tx: UnboundedSender<ApiTask>,
    ntfy_tx: UnboundedSender<Message>,
) -> Result<()> {
    let lua = Lua::new();
    let globals = lua.globals();

    let config = lua.load(config).set_name("config");

    lua.scope(|scope| {
        let log_trace_fn = scope.create_function(|_, s: String| {
            tracing::trace!(name: "lua", "{}", s);
            Ok(())
        })?;
        let log_debug_fn = scope.create_function(|_, s: String| {
            tracing::debug!(name: "lua", "{}", s);
            Ok(())
        })?;
        let log_info_fn = scope.create_function(|_, s: String| {
            tracing::info!(name: "lua", "{}", s);
            Ok(())
        })?;
        let log_warn_fn = scope.create_function(|_, s: String| {
            tracing::warn!(name: "lua", "{}", s);
            Ok(())
        })?;
        let log_error_fn = scope.create_function(|_, s: String| {
            tracing::error!(name: "lua", "{}", s);
            Ok(())
        })?;

        let log_tbl = lua.create_table_with_capacity(0, 5)?;
        log_tbl.set("trace", log_trace_fn)?;
        log_tbl.set("debug", log_debug_fn)?;
        log_tbl.set("info", log_info_fn)?;
        log_tbl.set("warn", log_warn_fn)?;
        log_tbl.set("error", log_error_fn)?;

        let ntfy_fn = scope.create_function_mut(|_, data: mlua::Value| {
            let data: NtfyMessage = lua.from_value(data)?;
            tracing::trace!(topic = data.topic, title = data.title, msg = data.message,"Sending Ntfy message");

            match ntfy_tx.send(Message::Ntfy(data)) {
                Ok(_) => Ok((true, mlua::Value::Nil)),
                Err(_) => {
                    let msg = lua.create_string("Failed to send message")?;
                    Ok((false, mlua::Value::String(msg)))
                }
            }
        })?;

        let set_api_task_fn = scope.create_function_mut(|_, data: mlua::Value| {
            let task: ApiTask = lua.from_value(data)?;
            tracing::trace!("Sending task request: id = {}, delay = {}", task.id, task.delay);

            match api_tx.send(task) {
                Ok(_) => Ok((true, mlua::Value::Nil)),
                Err(_) => {
                    let msg = lua.create_string("Failed to trigger task")?;
                    Ok((false, mlua::Value::String(msg)))
                }
            }
        })?;

        globals.set("log", log_tbl)?;
        globals.set("ntfy", ntfy_fn)?;
        globals.set("api_task", set_api_task_fn)?;

        config.exec()?;

        let event_fn: Function = match globals.get("on_event") {
            Ok(f) => f,
            Err(err) => match err {
                mlua::Error::FromLuaConversionError { from, to: _, message: _ } => {
                    let err = mlua::Error::runtime(format!("Global function 'on_event' not defined properly. Got value of type '{}'", from));
                    return Err(err);
                }
                err => return Err(err),
            },
        };

        // Main blocking loop. As long as we can receive events, this scope will stay active.
        while let Ok(event) = event_rx.recv() {
            match event {
                Event::Webhook(data) => {
                    tracing::trace!(id = data.topic, "Received webhook event");
                    let data = lua.to_value(&data)?;
                    event_fn.call(("webhook", data))?
                }
                Event::Api(data) => {
                    tracing::trace!(id = data.id, status = data.status, "Received api event");
                    let data = data.into_lua(&lua)?;
                    event_fn.call(("api", data))?
                }
                Event::Error(data) => {
                    tracing::trace!(id = data.id, message = data.message, "Received error event");
                    let data = lua.to_value(&data)?;
                    event_fn.call(("error", data))?
                }
            }
        }

        Ok(())
    }).map_err(|e| eyre::eyre!("{:?}", e))?;

    Ok(())
}
